const fs = require('fs');
const R = require('ramda');
const apiWrapper = require('./api-wrapper');
const tools = require('../utils/tools');
const profileHelper = require('../utils/profile-helper');
const optionValidator = require('../commands/option-validator');

// Public
module.exports = {
    createCommand: (commander) => {
        buildProfileNluCommand(commander);
        buildEvaluateNLUCommand(commander);
        buildListNLUEvaluationsCommand(commander);
        buildGetNLUEvaluationCommand(commander);
        buildGetNLUEvaluationResultsCommand(commander);
        buildCreateNLUAnnotationSetCommand(commander);
        buildDeleteNLUAnnotationSetCommand(commander);
        buildGetNLUAnnotationSetAnnotationsCommand(commander);
        buildGetNLUAnnotationSetPropertiesCommand(commander);
        buildListNLUAnnotationSetsCommand(commander);
        buildUpdateNLUAnnotationSetAnnotationsCommand(commander);
        buildUpdateNLUAnnotationSetPropertiesCommand(commander);
    }
};

// Private
function buildProfileNluCommand(commander) {
    commander
        .command('nlu-profile')
        .usage('<-s|--skill-id <skill-id>> <-l|--locale <locale>> <-u|--utterance <utterance>>'
            + ' [-g|--stage <stage>] [-t|--multiturn-token <multiturn-token>]'
            + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-g, --stage <stage>', 'skill stage (development or live). Defaults to development.')
        .option('-l, --locale <locale>', 'locale for the skill. Example: en-US.')
        .option('-u, --utterance <utterance>', 'utterance to be profiled')
        .option('-t, --multiturn-token <multiturn-token>', 'optional multiturn token for dialog')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('profile utterance with NLU')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateStage(options.stage, { required: false });
            optionValidator.validateUtterance(options.utterance);
            optionValidator.validateLocale(options.locale);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callProfileNlu(options.skillId, options.stage, options.locale, options.utterance,
            options.multiturnToken, profile, options.debug, (data) => {
                const response = tools.convertDataToJsonObject(data.body);
                if (response) {
                    console.log(JSON.stringify(response, null, 2));
                }
            });
    }
}

function buildEvaluateNLUCommand(commander) {
    commander
        .command('evaluate-nlu')
        .usage('<-s|--skill-id <skill-id>>  <-l|--locale <locale>>'
                + ' <-a|--annotation-id <annotation-id>>'
                + ' [-g|--stage <stage>] '
                + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-l, --locale <locale>', 'locale for the evaluation')
        .option('-a, --annotation-id <annotation-id>', 'annotationId for the evaluation')
        .option('-g, --stage <stage>', 'optional stage for the skill. Default is development')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('evaluate with NLU')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateStage(options.stage, { required: false });
            optionValidator.validateLocale(options.locale);
            optionValidator.validateAnnotationId(options.annotationId);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);

        apiWrapper.callEvaluateNLU(options.skillId, options.stage, options.locale,
            options.annotationId, profile, options.debug, (data) => {
                const evaluateResponse = tools.convertDataToJsonObject(data.body);
                if (!evaluateResponse) {
                    console.error('[Error]: Evaluate with NLU failed to get response.'
                    + ' Please run this command again with --debug for more details.');
                    process.exit(1);
                }
                console.log(JSON.stringify(evaluateResponse, null, 2));
            });
    }
}

function buildListNLUEvaluationsCommand(commander) {
    commander
        .command('list-nlu-evaluations')
        .usage('<-s|--skill-id <skill-id>> [-g|--stage <stage>] [-l|--locale <locale>]'
                + ' [--next-token <next-token>] [--max-results <max-results>]'
                + ' [-a|--annotation-id <annotation-id>][-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-g, --stage <stage>', 'stage filter for NLU evalution list')
        .option('-l, --locale <locale>', 'locale filter for NLU evalution list')
        .option('--next-token <next-token>', 'next token for NLU evalution list')
        .option('--max-results <max-results>', 'max results for NLU evaluation list')
        .option('-a, --annotation-id <annotation-id>', 'annotationId filter for NLU evalution list')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('list NLU evaluations for the skill')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateStage(options.stage, { required: false });
            optionValidator.validateLocale(options.locale, { required: false });
            optionValidator.validateAnnotationId(options.annotationId, { required: false });
            optionValidator.validateNextToken(options.nextToken, { required: false });
            optionValidator.validateMaxResults(options.maxResults, { required: false });
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callListNLUEvaluations(options.skillId, options.stage, options.locale,
            options.nextToken, options.maxResults, options.annotationId, profile, options.debug,
            (data) => {
                const response = tools.convertDataToJsonObject(data.body);
                if (response) {
                    const updatedResponse = {
                        ...removeLinksProperty(response),
                        evaluations: removeLinksFromJsonArrayItems(response.evaluations),
                    };
                    console.log(JSON.stringify(updatedResponse, null, 2));
                }
            });
    }
}

function buildGetNLUEvaluationCommand(commander) {
    commander
        .command('get-nlu-evaluation')
        .usage('<-s|--skill-id <skill-id>> <-e|--evaluation-id <evaluation-id>>'
                + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-e, --evaluation-id <evaluation-id>', 'evaluation id')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('get NLU evaluation')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateEvaluationId(options.evaluationId);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callGetNLUEvaluation(options.skillId, options.evaluationId,
            profile, options.debug, (data) => {
                const response = tools.convertDataToJsonObject(data.body);
                if (response) {
                    console.log(JSON.stringify(removeLinksProperty(response), null, 2));
                }
            });
    }
}

function buildGetNLUEvaluationResultsCommand(commander) {
    commander
        .command('get-nlu-evaluation-results')
        .usage('<-s|--skill-id <skill-id>> <-e|--evaluation-id <evaluation-id>>'
                + ' [--sort-field <sort-field>] [--test-case-status <test-case-status>]'
                + ' [--actual-intent-name <actual-intent-name>] [--expected-intent-name <expected-intent-name>]'
                + ' [--next-token <next-token>] [--max-results <max-results>]'
                + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-e, --evaluation-id <evaluation-id>', 'evaluation id for NLU evaluation')
        .option('--sort-field <sort-field>', 'sort field for NLU evalution results.'
                + ' Possible values: STATUS, ACTUAL_INTENT, EXPECTED_INTENT')
        .option('--test-case-status <test-case-status>', 'test case status filter for NLU evalution results.'
                + ' Possible values: PASSED or FAILED')
        .option('--expected-intent-name <expected-intent-name>', 'expected intent name filter for NLU evalution results.')
        .option('--actual-intent-name <actual-intent-name>', 'acutal intent name filter for NLU evalution results.')
        .option('--next-token <next-token>', 'next token for NLU evalution results.')
        .option('--max-results <max-results>', 'max results for NLU evaluation results.')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('get details of an evaluation')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateEvaluationId(options.evaluationId);
            optionValidator.validateSortField(options.sortField, { required: false });
            optionValidator.validateTestCaseStatus(options.testCaseStatus, { required: false });
            optionValidator.validateExpectedIntentName(options.expectedIntentName, { required: false });
            optionValidator.validateActualIntentName(options.actualIntentName, { required: false });
            optionValidator.validateNextToken(options.nextToken, { required: false });
            optionValidator.validateMaxResults(options.maxResults, { required: false });
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callGetNLUEvaluationResults(options.skillId, options.evaluationId,
            options.sortField, options.testCaseStatus, options.actualIntentName,
            options.expectedIntentName, options.nextToken, options.maxResults,
            profile, options.debug, (data) => {
                const response = tools.convertDataToJsonObject(data.body);
                if (response) {
                    console.log(JSON.stringify(removeLinksProperty(response), null, 2));
                }
            });
    }
}

function buildCreateNLUAnnotationSetCommand(commander) {
    commander
        .command('create-nlu-annotation-set')
        .usage('<-s|--skill-id <skill-id>> <-l|--locale <locale>> <-n|--annotation-set-name <annotation-set-name>>'
                + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-l, --locale <locale>', 'locale for the annotation set')
        .option('-n, --annotation-set-name <annotation-set-name>', 'annotation set name')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('creates NLU annotation set')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateLocale(options.locale);
            optionValidator.validateAnnotationSetName(options.annotationSetName);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callCreateNLUAnnotationSet(options.skillId, options.locale, options.annotationSetName, profile, options.debug, (data) => {
            const response = tools.convertDataToJsonObject(data.body);
            if (response) {
                console.log(JSON.stringify(response, null, 2));
            }
        });
    }
}

function buildDeleteNLUAnnotationSetCommand(commander) {
    commander
        .command('delete-nlu-annotation-set')
        .usage('<-s|--skill-id <skill-id>> <-a|--annotation-id <annotation-id>> [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-a, --annotation-id <annotation-id>', 'annotation-id for the annotation set')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('deletes NLU annotation set')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateAnnotationId(options.annotationId);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callDeleteNLUAnnotationSet(options.skillId, options.annotationId, profile, options.debug, () => {
            console.log('Annotation Set deleted successfully.');
        });
    }
}

function buildGetNLUAnnotationSetAnnotationsCommand(commander) {
    commander
        .command('get-nlu-annotation-set-annotations')
        .usage('<-s|--skill-id <skill-id>> <-a|--annotation-id <annotation-id>> <--annotations-format <annotations-format>>'
            + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-a, --annotation-id <annotation-id>', 'annotation-id for the annotation set')
        .option('--annotations-format <annotations-format>', 'format for the annotation set.'
                + ' Possible values: (application/json or text/csv)')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('get NLU annotation set annotations content')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateAnnotationId(options.annotationId);
            optionValidator.validateAnnotationsFormat(options.annotationsFormat);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callGetNLUAnnotationSetAnnotations(options.skillId, options.annotationId, options.annotationsFormat,
            profile, options.debug, (data) => {
                if (options.annotationsFormat === 'application/json') {
                    const response = tools.convertDataToJsonObject(data.body);
                    if (response) {
                        console.log(JSON.stringify(response, null, 2));
                    }
                } else {
                    console.log(data.body);
                }
            });
    }
}

function buildGetNLUAnnotationSetPropertiesCommand(commander) {
    commander
        .command('get-nlu-annotation-set-properties')
        .usage('<-s|--skill-id <skill-id>> <-a|--annotation-id <annotation-id>> [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-a, --annotation-id <annotation-id>', 'annotation-id for the annotation set')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('get NLU annotation set annotations content')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateAnnotationId(options.annotationId);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callGetNLUAnnotationSetProperties(options.skillId, options.annotationId, profile, options.debug, (data) => {
            const response = tools.convertDataToJsonObject(data.body);
            if (response) {
                console.log(JSON.stringify(response, null, 2));
            }
        });
    }
}

function buildListNLUAnnotationSetsCommand(commander) {
    commander
        .command('list-nlu-annotation-sets')
        .usage('<-s|--skill-id <skill-id>> [-l|--locale <locale>] [--next-token <next-token>] [--max-results <max-results>]'
            + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-l, --locale <locale>', 'locale filter for NLU annotation set list')
        .option('--next-token <next-token>', 'next token for NLU annotation set list')
        .option('--max-results <max-results>', 'max results for NLU eannotation set list')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('list NLU annotation sets for the skill')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateLocale(options.locale, { required: false });
            optionValidator.validateNextToken(options.nextToken, { required: false });
            optionValidator.validateMaxResults(options.maxResults, { required: false });
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callListNLUAnnotationSets(options.skillId, options.locale, options.nextToken, options.maxResults,
            profile, options.debug, (data) => {
                const response = tools.convertDataToJsonObject(data.body);
                if (response) {
                    console.log(JSON.stringify(removeLinksProperty(response), null, 2));
                }
            });
    }
}

function buildUpdateNLUAnnotationSetAnnotationsCommand(commander) {
    commander
        .command('update-nlu-annotation-set-annotations')
        .usage('<-s|--skill-id <skill-id>> <-a|--annotation-id <annotation-id>> <-f|--annotations-file <annotations-file>>'
            + ' <--annotations-format <annotations-format>> [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-a, --annotation-id <annotation-id>', 'annotation-id for the annotation set')
        .option('-f, --annotations-file <annotations-file>', 'annotation set content file')
        .option('--annotations-format <annotations-format>', 'file format of the annotation set content file.'
                + ' Possible values:  (application/json or text/csv)')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('updates NLU annotation set content')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateAnnotationId(options.annotationId);
            optionValidator.validateAnnotationsFile(options.annotationsFile);
            optionValidator.validateAnnotationsFormat(options.annotationsFormat);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);

        const annotationsContent = fs.readFileSync(options.annotationsFile, { encoding: 'UTF-8' });

        apiWrapper.callUpdateNLUAnnotationSetAnnotations(options.skillId, options.annotationId, annotationsContent, options.annotationsFormat,
            profile, options.debug, () => {
                console.log('Annotation Set content updated successfully.');
            });
    }
}

function buildUpdateNLUAnnotationSetPropertiesCommand(commander) {
    commander
        .command('update-nlu-annotation-set-properties')
        .usage('<-s|--skill-id <skill-id>> <-a|--annotation-id <annotation-id>> <-n|--annotation-set-name <annotation-set-name>>'
                + ' [-p|--profile <profile>] [--debug]')
        .option('-s, --skill-id <skill-id>', 'skill-id for the skill')
        .option('-a, --annotation-id <annotation-id>', 'annotation-id for the annotation set')
        .option('-n, --annotation-set-name <annotation-set-name>', 'annotation set name')
        .option('-p, --profile <profile>', 'optional ask cli profile')
        .option('--debug', 'optional ask cli debug mode')
        .description('updates NLU annotation set')
        .action(handle);

    function handle(options) {
        try {
            optionValidator.validateSkillId(options.skillId);
            optionValidator.validateAnnotationId(options.annotationId);
            optionValidator.validateAnnotationSetName(options.annotationSetName);
        } catch (err) {
            console.error(`[Error]: ${err}`);
            process.exit(1);
        }

        const profile = profileHelper.runtimeProfile(options.profile);
        apiWrapper.callUpdateNLUAnnotationSetProperties(options.skillId, options.annotationId,
            options.annotationSetName, profile, options.debug, () => {
                console.log('Annotation Set properties updated successfully.');
            });
    }
}

function removeLinksFromJsonArrayItems(inputJsonArray) {
    return R.map(R.omit(['_links'], R.__), inputJsonArray);
}

function removeLinksProperty(inputJson) {
    return R.omit(['_links'], inputJson);
}
