const fs = require('fs');
const path = require('path');
const tools = require('../utils/tools');
const gitClient = require('../clients/git-client');
const jsonfile = require('jsonfile');
const apiWrapper = require('../api/api-wrapper');
const displayUtil = require('../utils/display-utility');

module.exports = {
    getHostedSkillMetadata: getHostedSkillMetadata,
    clone: clone
}

/**
 * Setup the git repo, checkout, add .ask/ folder to .gitignore then add voicehub related fields to .ask/config
 * @param {Object} skillInfo returned from getHostedSkillMetadata(skillId)
 * @param {String} projectPath
 * @param {String} askConfigPath
 */
function clone(hostedSkillMetadata, projectPath, askConfigPath, profile, doDebug) {
    const verbosityOptions = {
        showCommand: doDebug ? true : false,
        showOutput: doDebug ? true : false
    }

    const spinner = new displayUtil.Spinner('Cloning Alexa Hosted Skill...');
    if (!doDebug) {
        spinner.start();
    }

    const repositoryUrl = hostedSkillMetadata.repository.url;
    addAHSFieldsToConfig(askConfigPath, profile);

    if (doDebug) {
        console.log(`- Setting up git repo...`);
    }
    setupGitRepo(projectPath, repositoryUrl, profile, verbosityOptions);
    if (doDebug) {
        console.log(`- Fetching git repo...`);
    }
    gitClient.fetchAll(projectPath, verbosityOptions);
    if (doDebug) {
        console.log('- Checking out dev branch...');
    }
    gitClient.checkoutBranch(projectPath, 'dev', verbosityOptions);
    if (doDebug) {
        console.log('- Setting up .gitignore...');
    }
    setupGitIgnore(projectPath, verbosityOptions);

    if (doDebug) {
        console.log(`- Git repo successfully setup`);
    }

    // Checkout all branches silently to make them visible when using git branch
    const branchesToCheckout = ['master', 'prod', 'dev'];
    branchesToCheckout.forEach((branch) => {
        gitClient.checkoutBranch(projectPath, branch, verbosityOptions)
    });

    if (!doDebug) {
        spinner.stop();
    }
}

/**
 * Setup a git repo, configure a local credentials helpers and add the origin
 * @param {*} projectPath
 * @param {*} repositoryUrl
 */
function setupGitRepo(projectPath, repositoryUrl, profile, verbosityOptions) {
    gitClient.init(projectPath, verbosityOptions);
    const credentialHelperPath = `ask util git-credentials-helper --profile ${profile} --operation $@`;
    gitClient.configureCredentialHelper(projectPath, credentialHelperPath, verbosityOptions);
    gitClient.addOrigin(projectPath, repositoryUrl, verbosityOptions);
}

/**
 * Create or append .ask/, hooks/ and lambda/node_modules to .gitignore then git add the .gitignore file
 * @param {*} projectPath
 */
function setupGitIgnore(projectPath, verbosityOptions) {
    const filesToIgnore = ['lambda/node_modules', '.ask/', 'hooks/'];
    const gitignorePath = path.join(projectPath, '.gitignore');
    if (fs.existsSync(gitignorePath) === false) {
        fs.writeFileSync(gitignorePath, `${filesToIgnore.join('\n')}`);
    } else {
        // check if "filesToIgnore" already there (no need to use a stream here, gitignore files are really small)
        const gitignoreFile = fs.readFileSync(gitignorePath);
        filesToIgnore.forEach(file => {
            if (gitignoreFile.toString().indexOf(file) === -1) {
                fs.appendFileSync(gitignorePath, `\n${file}`);
            }
        });
    }
    gitClient.add(projectPath, '.gitignore', verbosityOptions);
}

/**
 * Adds { alexaHosted: { isAlexaHostedSkill: true } } to .ask/config
 * @param {*} askConfigPath
 */
function addAHSFieldsToConfig(askConfigPath, profile) {
    const askConfig = jsonfile.readFileSync(askConfigPath);
    if (!askConfig) {
        throw('Please ensure current working directory is the root of your skill project that contains the project config in the hidden .ask folder.');
    }
    askConfig.alexaHosted = askConfig.alexaHosted || {};
    askConfig.alexaHosted.isAlexaHostedSkill = true;
    askConfig.alexaHosted.profile = profile;
    jsonfile.writeFileSync(askConfigPath, askConfig, { spaces: 2 });
}

/**
 * Gets skill info of a hosted skill
 * @param {*} skillId
 * @param {*} callback null if not hosted skill
 */
function getHostedSkillMetadata(skillId, profile, doDebug, callback) {
    apiWrapper.getHostedSkillMetadata(skillId, profile, doDebug, (data) => {
        if (data.statusCode === 200) {
            const skillInfo = tools.convertDataToJsonObject(data.body);
            callback(null, skillInfo);
        } else if (data.statusCode === 404) {
            callback(null, null)
        } else {
            const error = `Error while getting Alexa Hosted Skill Metadata: \n${data.body} \n(status code: ${data.statusCode})`;
            callback(error);
        }
    });
}
