const fs = require('fs');
const path = require('path');
const R = require('ramda');
const CONSTANTS = require('@src/utils/constants');

module.exports = {
    validateStage,
    validateLocale,
    validateUtterance,
    validateSkillId,
    validateNextToken,
    validateMaxResults,
    validateSortFieldForListModelVersions,
    validateSortDirectionForListModelVersions,
    validateFileFormat,
    validateEvaluationId,
    validateSortField,
    validateTestCaseStatus,
    validateExpectedIntentName,
    validateActualIntentName,
    validateFile,
    validateAnnotationId,
    validateAnnotationSetName,
    validateAnnotationsFormat,
    validateAnnotationsFile
};

const localeRegex = /^[a-z]{2}-[A-Z]{2}$/;
const stagesWithoutCert = [CONSTANTS.SKILL.STAGE.DEVELOPMENT, CONSTANTS.SKILL.STAGE.LIVE];
const stages = [stagesWithoutCert, CONSTANTS.SKILL.STAGE.CERTIFICATION];
// Sets the field that the sorting is applied to. The only valid value is version.
const validSortFieldForListModelVersions = ["version"];
// Sets the sorting direction of the versions returned in the response. Valid values are asc and desc. 
// When set to asc the versions are returned in ascending order, and when set to desc the versions are returned in descending order.
const validSortDirectionForListModelVersions = ["asc", "desc"];
const validFileFormatsForNLUEvaluation = ['application/json', 'text/csv'];
const validSortFieldForNLUEvaluation = ['STATUS', 'ACTUAL_INTENT', 'EXPECTED_INTENT'];
const validTestCaseStatusForNLUEvaluation = ['PASSED', 'FAILED'];

/**
 * General stage option validation logic
 * @param {String} stage
 * @param {Object} opts
 */
function validateStage(stage, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'stage';
    }

    validateWrapper(stage, opts, () => {
        validateString(stage);
        const allowCert = typeof opts.allowCert === 'undefined' ? false : opts.allowCert;
        const validStages = allowCert ? stages : stagesWithoutCert;
        const valid = validStages.indexOf(stage) !== -1;
        if (!valid) {
            throw (`Value must be in (${validStages.join(',')}).`);
        }
    });
}

/**
 * General locale option validation logic
 * @param {String} locale
 * @param {Object} opts
 */
function validateLocale(locale, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'locale';
    }

    validateWrapper(locale, opts, () => {
        validateString(locale);
        if (!localeRegex.test(locale)) {
            throw ('Invalid Format. Example: en-US.');
        }
    });
}

/**
 * General utterance option validation logic
 * @param {String} utterance
 * @param {Object} opts
 */
function validateUtterance(utterance, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'utterance';
    }

    validateWrapper(utterance, opts, () => {
        validateString(utterance);
    });
}

/**
 * General skill-id option validation logic
 * @param {String} skillId
 * @param {Object} opts
 */
function validateSkillId(skillId, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'skill-id';
    }

    validateWrapper(skillId, opts, () => {
        validateString(skillId);
    });
}

/**
 * General nextToken option validation logic
 * @param {String} nextToken
 * @param {Object} opts
 */
function validateNextToken(nextToken, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'next-token';
    }

    validateWrapper(nextToken, opts, () => {
        validateString(nextToken);
    });
}

/**
 * General maxResults option validation logic
 * @param {String} maxResults
 * @param {Object} opts
 */
function validateMaxResults(maxResults, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'max-results';
    }

    validateWrapper(maxResults, opts, () => {
        validateString(maxResults);
    });
}

/**
 * General sortField option validation logic
 * @param {String} sortField
 * @param {Object} opts
 */
function validateSortFieldForListModelVersions(sortField, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'sort-field';
    }

    validateWrapper(sortField, opts, () => {
        validateString(sortField);

        const valid = validSortFieldForListModelVersions.indexOf(sortField) !== -1;
        if (!valid) {
            throw (`Value must be in (${validSortFieldForListModelVersions.join(',')}).`);
        }
    });
}

/**
 * General sortDirection option validation logic
 * @param {String} sortDirection
 * @param {Object} opts
 */
function validateSortDirectionForListModelVersions(sortDirection, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'sort-direction';
    }

    validateWrapper(sortDirection, opts, () => {
        validateString(sortDirection);

        const valid = validSortDirectionForListModelVersions.indexOf(sortDirection) !== -1;
        if (!valid) {
            throw (`Value must be in (${validSortDirectionForListModelVersions.join(',')}).`);
        }
    });
}

/**
 * General annotationId option validation logic
 * @param {String} annotationId
 * @param {Object} opts
 */
function validateAnnotationId(annotationId, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'annotation-id';
    }

    validateWrapper(annotationId, opts, () => {
        validateString(annotationId);
    });
}

/**
 * General annotationName option validation logic
 * @param {String} annotationsName
 * @param {Object} opts
 */
function validateAnnotationSetName(annotationSetName, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'annotation-set-name';
    }

    validateWrapper(annotationSetName, opts, () => {
        validateString(annotationSetName);
    });
}

/**
 * General annotationsFormat option validation logic
 * @param {String} annotationsFormat
 * @param {Object} opts
 */
function validateAnnotationsFormat(annotationsFormat, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'annotations-format';
    }

    validateWrapper(annotationsFormat, opts, () => {
        validateString(annotationsFormat);

        const valid = validFileFormatsForNLUEvaluation.indexOf(annotationsFormat) !== -1;
        if (!valid) {
            throw (`Value must be in (${validFileFormatsForNLUEvaluation.join(', ')}).`);
        }
    });
}

/**
 * Validation logic for annotations file exists in file system.
 * @param {String} annotationsFile
 * @param {Object} opts
 */
function validateAnnotationsFile(annotationsFile, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'annotations-file';
    }
    validateWrapper(annotationsFile, opts, () => {
        validateString(annotationsFile);
        if (!fs.existsSync(annotationsFile)) {
            throw ('Annotations file does not exist. Please verify file path is correct.');
        }
    });
}

/**
 * General evaluation id option validation logic
 * @param {String} evaluationId
 * @param {Object} opts
 */
function validateEvaluationId(evaluationId, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'evaluation-id';
    }

    validateWrapper(evaluationId, opts, () => {
        validateString(evaluationId);
    });
}

/**
 * General file-format option validation logic
 * @param {String} fileFormat
 * @param {Object} opts
 */
function validateFileFormat(fileFormat, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'file-format';
    }

    validateWrapper(fileFormat, opts, () => {
        validateString(fileFormat);

        const valid = validFileFormatsForNLUEvaluation.indexOf(fileFormat) !== -1;
        if (!valid) {
            throw (`Value must be in (${validFileFormatsForNLUEvaluation.join(', ')}).`);
        }
    });
}

/**
 * General sortField option validation logic
 * @param {String} sortField
 * @param {Object} opts
 */
function validateSortField(sortField, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'sort-field';
    }

    validateWrapper(sortField, opts, () => {
        validateString(sortField);

        const valid = validSortFieldForNLUEvaluation.indexOf(sortField) !== -1;
        if (!valid) {
            throw (`Value must be in (${validSortFieldForNLUEvaluation.join(', ')}).`);
        }
    });
}

/**
 * General testCaseStatus option validation logic
 * @param {String} testCaseStatus
 * @param {Object} opts
 */
function validateTestCaseStatus(testCaseStatus, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'test-case-status';
    }

    validateWrapper(testCaseStatus, opts, () => {
        validateString(testCaseStatus);

        const valid = validTestCaseStatusForNLUEvaluation.indexOf(testCaseStatus) !== -1;
        if (!valid) {
            throw (`Value must be in (${validTestCaseStatusForNLUEvaluation.join(', ')}).`);
        }
    });
}

/**
 * General expectedIntentName option validation logic
 * @param {String} expectedIntentName
 * @param {Object} opts
 */
function validateExpectedIntentName(expectedIntentName, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'expected-intent-name';
    }

    validateWrapper(expectedIntentName, opts, () => {
        validateString(expectedIntentName);
    });
}

/**
 * General actualIntentName option validation logic
 * @param {String} actualIntentName
 * @param {Object} opts
 */
function validateActualIntentName(actualIntentName, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'actual-intent-name';
    }

    validateWrapper(actualIntentName, opts, () => {
        validateString(actualIntentName);
    });
}

/**
 * Validation logic for file exists in file system.
 * @param {String} file
 * @param {Object} opts
 */
function validateFile(file, opts = {}) {
    if (typeof opts.field === 'undefined') {
        opts.field = 'file';
    }
    validateWrapper(file, opts, () => {
        validateString(file);
        if (!fs.existsSync(file)) {
            throw ('File not exsist. Please verify file path is correct.');
        }
    });
}

function validateString(val) {
    if (!R.is(String, val)) {
        throw ('Must be a string.');
    }
    if (R.isEmpty(val.trim())) {
        throw ('Value must not be empty.');
    }
}

function validateWrapper(val, options = {}, callback) {
    const { field, required = true } = options;
    try {
        if (!R.isNil(val)) {
            if (callback) {
                callback();
            }
        } else if (required) {
            throw ('Field is required and must be set.');
        }
    } catch (err) {
        throw (`Please provide valid input for option: ${field}. ${err}`);
    }
}
