const { expect } = require('chai');
const sinon = require('sinon');
const http = require('http');
const proxyquire = require('proxyquire');
const portscanner = require('portscanner');

const ui = require('@src/commands/create-hosted-skill/ui');
const displayUtil = require('@src/utils/display-utility');

describe('Commands create-hosted-skill test - captcha-validator test', () => {
    const VENDOR_ID = 'VENDOR_ID';
    const CAPTCHA_URL = 'editor_url';
    const ENV_VAR_SIGNIN_HOST = process.env.ASK_LWA_AUTHORIZE_HOST;
    const LOGIN_URI = '?openid.ns=http%3A%2F%2Fspecs.openid.net%2Fauth%2F2.0' +
        '&openid.mode=checkid_setup' +
        '&openid.claimed_id=http%3A%2F%2Fspecs.openid.net%2Fauth%2F2.0%2Fidentifier_select' +
        '&openid.identity=http%3A%2F%2Fspecs.openid.net%2Fauth%2F2.0%2Fidentifier_select' +
        '&openid.assoc_handle=amzn_dante_us' +
        '&openid.return_to=' + CAPTCHA_URL +
        '%3Fvendor_id%3D' + VENDOR_ID +
        '%26redirect_url%3Dhttp%3A%2F%2F127.0.0.1%3A9090%2Fcaptcha' +
        '&openid.pape.max_auth_age=7200';

    describe('# solveCaptcha', () => {
        let captchaValidator;
        let stubOpn;
        let host;
        beforeEach(() => {
            host = 'https://www.amazon.com/ap/signin';
            const stubSpinner = sinon.stub(displayUtil, 'Spinner');
            stubSpinner.prototype.start = sinon.stub().returns(null);
            stubSpinner.prototype.stop = sinon.stub().returns(null);

            stubOpn = sinon.stub();
            captchaValidator = proxyquire('@src/commands/create-hosted-skill/captcha-validator', { 'open': stubOpn });
            sinon.stub(ui, 'getConfirmation');
            sinon.stub(process, 'exit');
            sinon.stub(portscanner, 'checkPortStatus');
        });

        it('| Solve captcha with --no-browser and confirm captcha validation is done', (done) => {
            // setup
            ui.getConfirmation.callsArgOnWith(0, {}, true);

            // call & verify
            captchaValidator.solveCaptcha(false, VENDOR_ID, CAPTCHA_URL, (error) => {
                expect(error).to.be.undefined;
                done();
            });
        });

        it('| Solve captcha with --no-browser and confirm captcha validation not done', (done) => {
            // setup
            ui.getConfirmation.callsArgOnWith(0, {}, false);

            // call & verify
            captchaValidator.solveCaptcha(false, VENDOR_ID, CAPTCHA_URL, (error) => {
                expect(process.exit.args[0][0]).equal(1);
                done();
            });
        });

        it(' | Solve captcha without --no-browser and start a local server using the signin host in env var', (done) => {
            // setup
            process.env.ASK_LWA_AUTHORIZE_HOST = 'https://development.amazon.com';
            host = 'https://development.amazon.com/ap/signin';
            const stubServer = {
                on: sinon.stub(),
                listen: sinon.stub()
            };
            sinon.stub(http, 'createServer').returns(stubServer);
            stubServer.on.onSecondCall().callsArgOnWith(1, null);
            portscanner.checkPortStatus.callsArgOnWith(1, {}, null, 'closed');

            // call & verify
            captchaValidator.solveCaptcha(true, VENDOR_ID, CAPTCHA_URL, (error) => {
                expect(stubOpn.args[0][0]).equal(host + LOGIN_URI);
                expect(http.createServer.calledOnce).to.be.true;
                expect(stubServer.on.getCall(0).args[0]).equal('connection');
                expect(stubServer.on.getCall(1).args[0]).equal('error');
                done();
            });
        });

        it(' | Solve captcha without --no-browser and start a local server using the signin host defined in constants', (done) => {
            // setup
            process.env.ASK_LWA_AUTHORIZE_HOST = '';
            const stubServer = {
                on: sinon.stub(),
                listen: sinon.stub()
            };
            sinon.stub(http, 'createServer').returns(stubServer);
            stubServer.on.onSecondCall().callsArgOnWith(1, null);
            portscanner.checkPortStatus.callsArgOnWith(1, {}, null, 'closed');

            // call & verify
            captchaValidator.solveCaptcha(true, VENDOR_ID, CAPTCHA_URL, (error) => {
                expect(stubOpn.args[0][0]).equal(host + LOGIN_URI);
                expect(http.createServer.calledOnce).to.be.true;
                expect(stubServer.on.getCall(0).args[0]).equal('connection');
                expect(stubServer.on.getCall(1).args[0]).equal('error');
                done();
            });
        });

        it(' | Solve captcha without --no-browser and failed to start localhost because port occupied', (done) => {
            // setup
            process.env.ASK_LWA_AUTHORIZE_HOST = '';
            const stubServer = {
                on: sinon.stub(),
                listen: sinon.stub()
            };
            sinon.stub(http, 'createServer').returns(stubServer);
            stubServer.on.onSecondCall().callsArgOnWith(1, null);
            portscanner.checkPortStatus.callsArgOnWith(1, {}, null, 'open');

            // call & verify
            captchaValidator.solveCaptcha(true, VENDOR_ID, CAPTCHA_URL, (error) => {
                expect(error.startsWith('[Warn]: 9090 port on localhost has been occupied')).to.be.true;
                expect(http.createServer.calledOnce).to.be.false;
                done();
            });
        });

        afterEach(() => {
            sinon.restore();
            process.env.ASK_LWA_AUTHORIZE_HOST = ENV_VAR_SIGNIN_HOST;
        });
    });
});
