const { expect } = require('chai');
const inquirer = require('inquirer');
const sinon = require('sinon');

const apiWrapper = require('@src/api/api-wrapper');
const clone = require('@src/clone/clone');
const handler = require('@src/commands/create-hosted-skill/handler');
const helper = require('@src/commands/create-hosted-skill/helper');
const ui = require('@src/commands/create-hosted-skill/ui');
const HostedSkillController = require('@src/controller/hosted-skill-controller/index');
const profileHelper = require('@src/utils/profile-helper');
const skillParser = require('@src/utils/skill-parser');

describe('Commands create-hosted-skill test - handler test', () => {
    const TEST_SKILL_NAME = 'skillName';
    const TEST_SKILL_ID = 'skillId';
    const TEST_FILTER_SKILL_NAME = 'filteredSkillName';
    const TEST_PROFILE = 'testProfile';
    const TEST_RUNTIME = 'nodejs8.10';
    const TEST_VENDOR_ID = 'vendorId';

    describe('# option checks', () => {
        let sandbox;
        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(console, 'error');
            sandbox.stub(process, 'exit');
        });

        it('| input operation is not a string, expect fatal error', () => {
            // setup
            const options = 'options';
            // call
            handler.handleOptions(options);
            // verify
            expect(console.error.args[0][0]).equal('[Error]: Invalid command. Please run "ask create-hosted-skill -h" for help.');
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| input runtime is invalid, expect fatal errror', () => {
            // setup
            const options = {
                runtime: 'invalidRuntime',
                skillName: TEST_SKILL_NAME,
                autoClone: 'false',
            };
            // call
            handler.handleOptions(options);
            // verify
            expect(console.error.args[0][0]).equal('[Error]: Invalid runtime option. Please run "ask create-hosted-skill -h" for help.');
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| missing bundle options, expect fatal errror', () => {
            // setup
            const options = {
                skillName: TEST_SKILL_NAME,
                autoClone: 'false',
            };
            // call
            handler.handleOptions(options);
            // verify
            expect(console.error.args[0][0]).equal('[Error]: Invalid command usage. --skill-name, '
                + '--runtime and --auto-clone must be either be set together or all absent.');
            expect(process.exit.args[0][0]).equal(1);
        });

        afterEach(() => {
            console.error.restore();
            process.exit.restore();
            sandbox.restore();
        });
    });

    describe('# UI error (getSkillName and getRuntime)', () => {
        let sandbox;
        const options = {};

        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(profileHelper, 'runtimeProfile');
            sandbox.stub(skillParser, 'filterUserSkillName');
            sandbox.stub(ui, 'getSkillName');
            sandbox.stub(ui, 'getRunTime');
            sandbox.stub(console, 'error');
            sandbox.stub(process, 'exit');
            sandbox.stub(profileHelper, 'resolveVendorId');
            sandbox.stub(helper, 'validateUserQualification');

            profileHelper.runtimeProfile.withArgs(TEST_PROFILE).returns(TEST_PROFILE);
            profileHelper.resolveVendorId.withArgs(TEST_PROFILE).returns(TEST_VENDOR_ID);
            skillParser.filterUserSkillName.withArgs(TEST_SKILL_NAME).returns(TEST_FILTER_SKILL_NAME);
            helper.validateUserQualification.callsArgOnWith(3, {}, null);
        });

        it('| ui getSkillName returns error, expect console error', () => {
            // setup
            const GET_SKILL_NAME_ERROR = 'getSkillNameError';
            ui.getSkillName.callsArgOnWith(1, {}, GET_SKILL_NAME_ERROR, null);
            // call
            handler.handleOptions(options);
            // verify
            expect(console.error.args[0][0]).equal(GET_SKILL_NAME_ERROR);
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| ui getRunTime returns error, expect console error', () => {
            // setup
            const GET_RUNTIME_ERROR = 'getRunTimeError';
            ui.getSkillName.callsArgOnWith(1, {}, null, TEST_SKILL_NAME);
            ui.getRunTime.callsArgOnWith(1, {}, GET_RUNTIME_ERROR, null);

            // call
            handler.handleOptions(options);
            // verify

            expect(console.error.args[0][0]).equal(GET_RUNTIME_ERROR);
            expect(process.exit.args[0][0]).equal(1);
        });

        afterEach(() => {
            sandbox.restore();
        });
    });

    describe('# create skill process error', () => {
        let sandbox;
        const options = {};

        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(apiWrapper, 'callDeleteSkill');
            sandbox.stub(profileHelper, 'runtimeProfile');
            sandbox.stub(profileHelper, 'resolveVendorId');
            sandbox.stub(skillParser, 'filterUserSkillName');
            sandbox.stub(console, 'error');
            sandbox.stub(console, 'log');
            sandbox.stub(process, 'exit');
            sandbox.stub(ui, 'getSkillName');
            sandbox.stub(ui, 'getRunTime');
            sandbox.stub(helper, 'validateUserQualification');
            sandbox.stub(HostedSkillController.prototype, 'createSkill');
            sandbox.stub(HostedSkillController.prototype, 'pollingSkillStatus');

            ui.getSkillName.callsArgOnWith(1, {}, null, TEST_SKILL_NAME);
            ui.getRunTime.callsArgOnWith(1, {}, null, TEST_RUNTIME);
            helper.validateUserQualification.callsArgOnWith(3, {}, null);

            profileHelper.resolveVendorId.withArgs(TEST_PROFILE).returns(TEST_VENDOR_ID);
            profileHelper.runtimeProfile.withArgs(TEST_PROFILE).returns(TEST_PROFILE);
            skillParser.filterUserSkillName.withArgs(TEST_SKILL_NAME).returns(TEST_FILTER_SKILL_NAME);
        });

        it('| validateUserQualification error', () => {
            // setup
            const VALIDATE_USER_ERROR = "validateUserQualificationError";
            helper.validateUserQualification.callsArgOnWith(3, {}, VALIDATE_USER_ERROR);

            //call
            handler.handleOptions(options);

            //verify
            expect(console.error.args[0][0]).equal(VALIDATE_USER_ERROR);
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| HostedSkillController create skill error', () => {
            // setup
            const CREATE_SKILL_ERROR = 'createSkillError';
            HostedSkillController.prototype.createSkill.callsArgOnWith(2, {}, CREATE_SKILL_ERROR, null);
            // call
            handler.handleOptions(options);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.error.args[0][0]).equal(CREATE_SKILL_ERROR);
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| HostedSkillController polling skill status method error', () => {
            // setup
            HostedSkillController.prototype.createSkill.callsArgOnWith(2, {}, null, TEST_SKILL_ID);
            HostedSkillController.prototype.pollingSkillStatus.callsArgOnWith(2, {}, 'error', null);
            apiWrapper.callDeleteSkill.callsArgOnWith(3, {}, null);
            // call
            handler.handleOptions(options);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.error.args[0][0]).equal('error');
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| HostedSkillController polling skill status response is invalid', () => {
            // setup
            const POLL_SKILL_STATUS_ERROR = '[Fatal]: Response from the creation of hosted skill is not valid.';
            HostedSkillController.prototype.createSkill.callsArgOnWith(2, {}, null, TEST_SKILL_ID);
            HostedSkillController.prototype.pollingSkillStatus.callsArgOnWith(2, {}, null, null);
            apiWrapper.callDeleteSkill.callsArgOnWith(3, {}, null);
            // call
            handler.handleOptions(options);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.error.args[0][0]).equal(POLL_SKILL_STATUS_ERROR);
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| HostedSkillController polling skill status response: in progress when timeout', () => {
            // setup
            const POLL_SKILL_STATUS_ERROR = '[Error]: Timeout when checking the status of hosted-skill creation. Please try again.';
            const response = {
                manifest: { status: 'SUCCEEDED' },
                interactionModel: { status: 'IN_PROGRESS' },
                provisioning: { status: 'SUCCEEDED' },
            };

            HostedSkillController.prototype.createSkill.callsArgOnWith(2, {}, null, TEST_SKILL_ID);
            HostedSkillController.prototype.pollingSkillStatus.callsArgOnWith(2, {}, null, response);
            apiWrapper.callDeleteSkill.callsArgOnWith(3, {}, null);

            // call
            handler.handleOptions(options);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.error.args[0][0]).equal(POLL_SKILL_STATUS_ERROR);
            expect(process.exit.args[0][0]).equal(1);
        });

        it('| HostedSkillController polling skill status response: failed', () => {
            // setup
            const POLL_SKILL_STATUS_ERROR = '[Error]: Failed to create hosted skill for the following reason:';
            const response = {
                manifest: { status: 'FAILED' },
                interactionModel: { status: 'FAILED' },
                provisioning: { status: 'FAILED' },
            };

            HostedSkillController.prototype.createSkill.callsArgOnWith(2, {}, null, TEST_SKILL_ID);
            HostedSkillController.prototype.pollingSkillStatus.callsArgOnWith(2, {}, null, response);
            apiWrapper.callDeleteSkill.callsArgOnWith(3, {}, null);

            // call
            handler.handleOptions(options);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.error.args[0][0]).equal(POLL_SKILL_STATUS_ERROR);
            expect(console.error.args[1][0]).equal('Skill provisioning step failed');
            expect(console.error.args[2][0]).equal('Skill interaction model building step failed');
            expect(console.error.args[3][0]).equal('Skill manifest building step failed');
            expect(process.exit.args[0][0]).equal(1);
        });

        afterEach(() => {
            sandbox.restore();
        });
    });

    describe('# create and clone hosted skill process', () => {
        let sandbox;
        const options = {};

        beforeEach(() => {
            sandbox = sinon.createSandbox();
            sandbox.stub(inquirer, 'prompt');
            inquirer.prompt.resolves({ isAllowed: 'true' });
            sandbox.stub(profileHelper, 'runtimeProfile');
            sandbox.stub(profileHelper, 'resolveVendorId');
            sandbox.stub(skillParser, 'filterUserSkillName');
            sandbox.stub(ui, 'getSkillName');
            sandbox.stub(ui, 'getRunTime');
            sandbox.stub(console, 'log');
            sandbox.stub(console, 'error');
            sandbox.stub(process, 'exit');
            sandbox.stub(HostedSkillController.prototype, 'createSkill');
            sandbox.stub(HostedSkillController.prototype, 'pollingSkillStatus');
            sandbox.stub(clone, 'singleSkillCloneProcess');
            sandbox.stub(helper, 'validateUserQualification');

            ui.getSkillName.callsArgOnWith(1, {}, null, TEST_SKILL_NAME);
            ui.getRunTime.callsArgOnWith(1, {}, null, TEST_RUNTIME);
            helper.validateUserQualification.callsArgOnWith(3, {}, null);

            profileHelper.runtimeProfile.withArgs(TEST_PROFILE).returns(TEST_PROFILE);
            profileHelper.resolveVendorId.withArgs(TEST_PROFILE).returns(TEST_VENDOR_ID);
            skillParser.filterUserSkillName.withArgs(TEST_SKILL_NAME).returns(TEST_FILTER_SKILL_NAME);
            HostedSkillController.prototype.createSkill.callsArgOnWith(2, {}, null, TEST_SKILL_ID);

            const response = {
                manifest: { status: 'SUCCEEDED' },
                interactionModel: { status: 'SUCCEEDED' },
                provisioning: { status: 'SUCCEEDED' },
            };
            HostedSkillController.prototype.pollingSkillStatus.callsArgOnWith(2, {}, null, response);
        });

        it('| Create skill and clone skill success', () => {
            // setup
            const CREATE_SKILL_SUCCESS = 'Hosted skill provisioning finished.';
            clone.singleSkillCloneProcess.callsArgOnWith(3, {}, null);
            // call
            handler.handleOptions(options);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.log.args[1][0]).equal(CREATE_SKILL_SUCCESS);
        });

        it('| Create skill but ui.getAutoClone fails with error', () => {
            // setup
            sandbox.stub(ui, 'getAutoClone');
            ui.getAutoClone.callsArgOnWith(1, {}, 'error');
            const optionsWithNoClone = {
                autoClone: 'false',
                skillName: TEST_SKILL_NAME,
                runtime: TEST_RUNTIME,
            };
            // call
            handler.handleOptions(optionsWithNoClone);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.error.args[0][0]).equal('error');
            expect(clone.singleSkillCloneProcess.called).equal(false);
        });

        it('| Create skill without auto clone skill', () => {
            // setup
            sandbox.stub(ui, 'getAutoClone');
            ui.getAutoClone.callsArgOnWith(1, {}, null, false);
            const optionsWithNoClone = {
                autoClone: 'false',
                skillName: TEST_SKILL_NAME,
                runtime: TEST_RUNTIME,
            };
            const CREATE_SKILL_SUCCESS = 'Hosted skill provisioning finished.';
            const NO_CLONE_MESSAGE = `Hosted skill created successfully. Skill-Id: ${TEST_SKILL_ID}`;
            // call
            handler.handleOptions(optionsWithNoClone);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(console.log.args[1][0]).equal(CREATE_SKILL_SUCCESS);
            expect(console.log.args[2][0]).equal(NO_CLONE_MESSAGE);
            expect(clone.singleSkillCloneProcess.called).equal(false);
        });

        it('| Create skill with auto clone', () => {
            // setup
            sandbox.stub(ui, 'getAutoClone');
            ui.getAutoClone.callsArgOnWith(1, {}, null, true);
            const optionsWithNoClone = {
                autoClone: 'true',
                skillName: TEST_SKILL_NAME,
                runtime: TEST_RUNTIME,
            };
            // call
            handler.handleOptions(optionsWithNoClone);
            // verify
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
            expect(clone.singleSkillCloneProcess.calledOnce).equal(true);
        });

        it('| Create skill with auto clone but clone error out', () => {
            // setup
            sandbox.stub(ui, 'getAutoClone');
            ui.getAutoClone.callsArgOnWith(1, {}, null, true);
            const optionsWithNoClone = {
                autoClone: 'true',
                skillName: TEST_SKILL_NAME,
                runtime: TEST_RUNTIME,
                profile: TEST_PROFILE
            };

            // call
            handler.handleOptions(optionsWithNoClone);
            // verify
            expect(clone.singleSkillCloneProcess.args[0][0]).equal(TEST_SKILL_ID);
            expect(clone.singleSkillCloneProcess.args[0][1]).equal(TEST_PROFILE);
            expect(console.log.args[0][0]).equal('-------------------- Create Alexa Hosted Skill Project --------------------');
        });

        afterEach(() => {
            sandbox.restore();
        });
    });
});
