const { expect } = require('chai');
const sinon = require('sinon');
const helper = require('@src/commands/create-hosted-skill/helper');
const captchaValidator = require('@src/commands/create-hosted-skill/captcha-validator');

describe('Commands create-hosted-skill test - helper test', () => {
    const PERMISSION_TYPE = 'newSkill';
    const PERMISSION_ALLOWED = 'ALLOWED';
    const PERMISSION_RATE_EXCEEDED = 'RATE_EXCEEDED';
    const PERMISSION_RESOURCE_LIMIT_EXCEEDED = 'RESOURCE_LIMIT_EXCEEDED';
    const PERMISSION_NEW_USER_REGISTRATION_REQUIRED = 'NEW_USER_REGISTRATION_REQUIRED';
    const VENDOR_ID = 'VENDOR_ID';

    describe('# checkPermission', () => {
        const options = {};
        let controller;

        beforeEach(() => {
            sinon.stub(captchaValidator, 'solveCaptcha');
            controller = {
                getHostedSkillPermission: sinon.stub()
            };
        });

        it('| Check permission returns error', (done) => {
            // setup
            const CHECK_PERMISSION_ERROR = 'getHostedSkillPermissionError';
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, CHECK_PERMISSION_ERROR, null);

            // call & verify
            helper.checkPermission(options, controller, false,(error, actionUrl) => {
                expect(error).equal(CHECK_PERMISSION_ERROR);
                expect(actionUrl).to.be.undefined;
                done();
            });
        });

        it('| Check permission returns invalid response', (done) => {
            // setup
            const CHECK_PERMISSION_INVALID_RESPONSE = '[Fatal]: Response from getHostedSkillPermission is not valid.';
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, null);

            // call & verify
            helper.checkPermission(options, controller, false,(error, actionUrl) => {
                expect(error).equal(CHECK_PERMISSION_INVALID_RESPONSE);
                expect(actionUrl).to.be.undefined;
                done();
            });
        });

        it('| Check permission returns RESOURCE_LIMIT_EXCEEDED', (done) => {
            // setup
            const RESOURCE_LIMIT_EXCEEDED_ERROR = '[Error]: Your account has reached the maximum limit for Alexa hosted skill.';
            const permissionInfo = {
                status: PERMISSION_RESOURCE_LIMIT_EXCEEDED,
                permission: PERMISSION_TYPE
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);

            // call & verify
            helper.checkPermission(options, controller, false,(error, actionUrl) => {
                expect(error.startsWith(RESOURCE_LIMIT_EXCEEDED_ERROR)).to.be.true;
                expect(actionUrl).to.be.undefined;
                done();
            });
        });

        it('| Check permission returns RATE_EXCEEDED', (done) => {
            // setup
            const RATE_EXCEEDED_ERROR = '[Error]: Your hosted skills account is limited to a certain number of new hosted skills per minute.';
            const permissionInfo = {
                status: PERMISSION_RATE_EXCEEDED,
                permission: PERMISSION_TYPE
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);

            // call & verify
            helper.checkPermission(options, controller, false,(error, actionUrl) => {
                expect(error.startsWith(RATE_EXCEEDED_ERROR)).to.be.true;
                expect(actionUrl).to.be.undefined;
                done();
            });
        });

        it('| Check permission returns NEW_USER_REGISTRATION_REQUIRED', (done) => {
            // setup
            const permissionInfo = {
                status: PERMISSION_NEW_USER_REGISTRATION_REQUIRED,
                permission: PERMISSION_TYPE,
                actionUrl: 'captchaValidationUrl'
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);

            // call & verify
            helper.checkPermission(options, controller, false,(error, actionUrl) => {
                expect(error).to.be.null;
                expect(actionUrl).equal(permissionInfo.actionUrl);
                done();
            });
        });

        it('| Recheck permission returns NEW_USER_REGISTRATION_REQUIRED', (done) => {
            // setup
            const permissionInfo = {
                status: PERMISSION_NEW_USER_REGISTRATION_REQUIRED,
                permission: PERMISSION_TYPE,
                actionUrl: 'captchaValidationUrl'
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);

            // call & verify
            helper.checkPermission(options, controller, true,(error, actionUrl) => {
                expect(error).equal('[Error]: Failed to validate CAPTCHA. Please try again.');
                expect(actionUrl).to.be.undefined;
                done();
            });
        });

        it('| Check permission returns ALLOWED', (done) => {
            // setup
            const permissionInfo = {
                status: PERMISSION_ALLOWED,
                permission: PERMISSION_TYPE,
                actionUrl: ''
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);

            // call & verify
            helper.checkPermission(options, controller, false,(error, actionUrl) => {
                expect(error).to.be.undefined;
                expect(actionUrl).to.be.undefined;
                done();
            });
        });

        afterEach(() => {
            sinon.restore();
        });
    });

    describe('# validateUserQualification', () => {
        const options = {};
        let controller;

        beforeEach(() => {
            sinon.stub(captchaValidator, 'solveCaptcha');
            controller = {
                getHostedSkillPermission: sinon.stub()
            };
        });

        it('| CheckPermission failed', (done) => {
            // setup
            const CHECK_PERMISSION_ERROR = 'getHostedSkillPermissionError';
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, CHECK_PERMISSION_ERROR, null);

            // call & verify
            helper.validateUserQualification(options, VENDOR_ID, controller, (error) => {
                expect(error).equal(CHECK_PERMISSION_ERROR);
                done();
            });
        });

        it('| No action url returned', (done) => {
            // setup
            const permissionInfo = {
                status: PERMISSION_ALLOWED,
                permission: PERMISSION_TYPE,
                actionUrl: ''
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);

            // call & verify
            helper.validateUserQualification(options, VENDOR_ID, controller, (error) => {
                expect(error).to.be.undefined;
                done();
            });
        });

        it('| Captcha validation failed', (done) => {
            // setup
            const CAPTCHA_VALIDATE_ERROR = 'captchaValidateError';
            const permissionInfo = {
                status: PERMISSION_NEW_USER_REGISTRATION_REQUIRED,
                permission: PERMISSION_TYPE,
                actionUrl: 'captchaValidationUrl'
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);
            captchaValidator.solveCaptcha.callsArgOnWith(3, {}, CAPTCHA_VALIDATE_ERROR);

            // call & verify
            helper.validateUserQualification(options, VENDOR_ID, controller, (error) => {
                expect(error).equal(CAPTCHA_VALIDATE_ERROR);
                done();
            });
        });

        it('| Captcha validation failed', (done) => {
            // setup
            const CAPTCHA_VALIDATE_ERROR = 'captchaValidateError';
            const permissionInfo = {
                status: PERMISSION_NEW_USER_REGISTRATION_REQUIRED,
                permission: PERMISSION_TYPE,
                actionUrl: 'captchaValidationUrl'
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);
            captchaValidator.solveCaptcha.callsArgOnWith(3, {}, CAPTCHA_VALIDATE_ERROR);

            // call & verify
            helper.validateUserQualification(options, VENDOR_ID, controller, (error) => {
                expect(error).equal(CAPTCHA_VALIDATE_ERROR);
                done();
            });
        });

        it('| Recheck permission failed', (done) => {
            // setup
            const RECHECK_PERMISSION_FAILED = '[Error]: Failed to validate CAPTCHA. Please try again.';
            const permissionInfo = {
                status: PERMISSION_NEW_USER_REGISTRATION_REQUIRED,
                permission: PERMISSION_TYPE,
                actionUrl: 'captchaValidationUrl'
            };
            controller.getHostedSkillPermission.callsArgOnWith(2, {}, null, permissionInfo);
            captchaValidator.solveCaptcha.callsArgOnWith(3, {}, null);

            // call & verify
            helper.validateUserQualification(options, VENDOR_ID, controller, (error) => {
                expect(error).equal(RECHECK_PERMISSION_FAILED);
                done();
            });
        });

        it('| Captcha validation succeeded', (done) => {
            // setup
            const permissionInfoA = {
                status: PERMISSION_NEW_USER_REGISTRATION_REQUIRED,
                permission: PERMISSION_TYPE,
                actionUrl: 'captchaValidationUrl'
            };
            const permissionInfoB = {
                status: PERMISSION_ALLOWED,
                permission: PERMISSION_TYPE,
                actionUrl: ''
            };
            controller.getHostedSkillPermission.onFirstCall().callsArgOnWith(2, {}, null, permissionInfoA);
            controller.getHostedSkillPermission.onSecondCall().callsArgOnWith(2, {}, null, permissionInfoB);
            captchaValidator.solveCaptcha.callsArgOnWith(3, {}, null);

            // call & verify
            helper.validateUserQualification(options, VENDOR_ID, controller, (error) => {
                expect(error).to.be.undefined;
                done();
            });
        });

        afterEach(() => {
            sinon.restore();
        });
    });
});
