'use strict';

const expect = require('chai').expect;
const sinon = require('sinon');
const fs = require('fs');
const CONSTANTS = require('@src/utils/constants');
const optionValidator = require('@src/commands/option-validator');
const validSortFieldForListModelVersions = ["version"];
const validSortDirectionForListModelVersions = ["asc", "desc"];
const validFileFormatsForNLUEvaluation = ["application/json", "text/csv"];
const validSortFieldForNLUEvaluation = ["STATUS", "ACTUAL_INTENT", "EXPECTED_INTENT"];
const validTestCaseStatusForNLUEvaluation = ["PASSED", "FAILED"];

describe('option validation testing', () => {
	let sandbox;

    beforeEach(() => {
        sandbox = sinon.sandbox.create();
        sandbox.stub(fs, 'existsSync');
    });

    afterEach(() => {
        sandbox.restore();
    });

    describe('# validate stage', () => {
        it('| should match development stage', () => {
            expect(() => optionValidator.validateStage(CONSTANTS.SKILL.STAGE.DEVELOPMENT)).to.not.throw();
        });

        it('| should match live stage', () => {
            expect(() => optionValidator.validateStage(CONSTANTS.SKILL.STAGE.LIVE)).to.not.throw();
        });

        it('| should expect strict casing', () => {
            expect(() => optionValidator.validateStage(CONSTANTS.SKILL.STAGE.DEVELOPMENT.toUpperCase()))
                .to.throw('Please provide valid input for option: stage. Value must be in (development,live).');
        });

        it('| should match cert when allowed', () => {
            expect(() => optionValidator.validateStage(CONSTANTS.SKILL.STAGE.CERTIFICATION, {allowCert: true}))
                .to.not.throw();
        });

        it('| should not match cert when not allowed', () => {
            expect(() => optionValidator.validateStage(CONSTANTS.SKILL.STAGE.CERTIFICATION, {allowCert: false}))
                .to.throw('Please provide valid input for option: stage. Value must be in (development,live).');
        });

        validatorTests(optionValidator.validateStage, 'stage');
    });

    describe('# validate locale', () => {
        it('| should allow valid format', () => {
            expect(() => optionValidator.validateLocale('en-US')).to.not.throw();
        });

        it('| should be case sensitive', () => {
            expect(() => optionValidator.validateLocale('En-uS'))
                .to.throw('Please provide valid input for option: locale. Invalid Format. Example: en-US.');
        });

        it('| should not allow language only', () => {
            expect(() => optionValidator.validateLocale('en'))
                .to.throw('Please provide valid input for option: locale. Invalid Format. Example: en-US.');
        });

        it('| should not allow invalid format', () => {
            expect(() => optionValidator.validateLocale('en_US'))
                .to.throw('Please provide valid input for option: locale. Invalid Format. Example: en-US.');
        });

        it('| should not allow whitespace', () => {
            expect(() => optionValidator.validateLocale(' en-us'))
                .to.throw('Please provide valid input for option: locale. Invalid Format. Example: en-US.');
            expect(() => optionValidator.validateLocale('en-us '))
                .to.throw('Please provide valid input for option: locale. Invalid Format. Example: en-US.');
        });

        validatorTests(optionValidator.validateLocale, 'locale');
    });

    describe('# validate utterance', () => {
        it('| should allow valid format', () => {
            expect(() => optionValidator.validateUtterance('I am an utterance')).to.not.throw();
        });

        validatorTests(optionValidator.validateUtterance, 'utterance');
    });

    describe('# validate skillId', () => {
        it('| should allow valid format', () => {
            expect(() => optionValidator.validateSkillId('amz1.ask.something')).to.not.throw();
        });

        validatorTests(optionValidator.validateSkillId, 'skill-id');
    });

    describe('# validate nextToken', () => {
         it('| should allow valid format', () => {
             expect(() => optionValidator.validateNextToken('nextToken')).to.not.throw();
         });
 
         validatorTests(optionValidator.validateNextToken, 'next-token');
     });
 
     describe('# validate maxResults', () => {
         it('| should allow valid format', () => {
             expect(() => optionValidator.validateMaxResults('maxResults')).to.not.throw();
         });
 
         validatorTests(optionValidator.validateMaxResults, 'max-results');
     });

    describe('# validate sortFieldForListModelVersions', () => {
        it('| should allow valid format', () => {
            validSortFieldForListModelVersions.forEach((sortField) => {
                expect(() => optionValidator.validateSortFieldForListModelVersions(sortField)).to.not.throw();
            });
        });

        it('| should expect strict casing', () => {
            validSortFieldForListModelVersions.forEach((sortField) => {
                expect(() => optionValidator.validateSortFieldForListModelVersions(sortField.toUpperCase()))
                    .to.throw('Please provide valid input for option: sort-field. Value must be in (version).');
            });
        });

        it('| should allow invalid format', () => {
            expect(() => optionValidator.validateSortFieldForListModelVersions('invalid-sort-field'))
                .to.throw('Please provide valid input for option: sort-field. Value must be in (version).');
        });

        validatorTests(optionValidator.validateSortFieldForListModelVersions, 'sort-field');
    });

    describe('# validate sortDirectionForListModelVersions', () => {
        it('| should allow valid format', () => {
            validSortDirectionForListModelVersions.forEach((sortField) => {
                expect(() => optionValidator.validateSortDirectionForListModelVersions(sortField)).to.not.throw();
            });
        });

        it('| should expect strict casing', () => {
            validSortDirectionForListModelVersions.forEach((sortField) => {
                expect(() => optionValidator.validateSortDirectionForListModelVersions(sortField.toUpperCase()))
                    .to.throw('Please provide valid input for option: sort-direction. Value must be in (asc,desc).');
            });
        });

        it('| should allow invalid format', () => {
            expect(() => optionValidator.validateSortDirectionForListModelVersions('invalid-sort-direction'))
                .to.throw('Please provide valid input for option: sort-direction. Value must be in (asc,desc).');
        });

        validatorTests(optionValidator.validateSortDirectionForListModelVersions, 'sort-direction');
    });

    describe('# validate annotationId', () => {
        it('| should allow annotationId', () => {
            expect(() => optionValidator.validateAnnotationId('annotationId')).to.not.throw();
        });

        validatorTests(optionValidator.validateAnnotationId, 'annotation-id');
    });

    describe('# validate validateAnnotationSetName', () => {
        it('| should allow valid annotationSetName', () => {
            expect(() => optionValidator.validateAnnotationSetName('annotationSetName')).to.not.throw();
        });

        validatorTests(optionValidator.validateAnnotationSetName, 'annotation-set-name');
    });

    describe('# validate annotationsFormat', () => {
        it('| should allow valid format', () => {
            for (var i = 0; i < validFileFormatsForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateAnnotationsFormat(validFileFormatsForNLUEvaluation[i])).to.not.throw();
            }
        });

        it('| should expect strict casing', () => {
            for (var i = 0; i < validFileFormatsForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateAnnotationsFormat(validFileFormatsForNLUEvaluation[i].toUpperCase()))
                .to.throw('Please provide valid input for option: annotations-format.'
                        + ' Value must be in (application/json, text/csv).');
            }
        });

        it('| should allow invalid format', () => {
            expect(() => optionValidator.validateAnnotationsFormat('invalid-annotations-format'))
            .to.throw('Please provide valid input for option: annotations-format.'
                    + ' Value must be in (application/json, text/csv).');
        });

        validatorTests(optionValidator.validateAnnotationsFormat, 'annotations-format');
    });

    describe('# validate annotations file exists', () => {
        it('| should allow valid file path', () => {
            fs.existsSync.returns(true);
            expect(() => optionValidator.validateAnnotationsFile('file-path')).to.not.throw();
        });

        it('| should not allow non existent file path', () => {
            fs.existsSync.returns(false);
            expect(() => optionValidator.validateAnnotationsFile('file-path'))
                .to.throw('Annotations file does not exist. Please verify file path is correct.');
        });

        validatorTests(optionValidator.validateAnnotationsFile, 'annotations-file');
    });

    describe('# validate evaluationId', () => {
        it('| should allow valid format', () => {
            expect(() => optionValidator.validateEvaluationId('Evaluation-ID')).to.not.throw();
        });

        validatorTests(optionValidator.validateEvaluationId, 'evaluation-id');
    });

    describe('# validate fileFormat', () => {
        it('| should allow valid format', () => {
            for (var i = 0; i < validFileFormatsForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateFileFormat(validFileFormatsForNLUEvaluation[i])).to.not.throw();
            }
        });

        it('| should expect strict casing', () => {
            for (var i = 0; i < validFileFormatsForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateFileFormat(validFileFormatsForNLUEvaluation[i].toUpperCase()))
                .to.throw('Please provide valid input for option: file-format.'
                        + ' Value must be in (application/json, text/csv).');
            }
        });

        it('| should allow invalid format', () => {
            expect(() => optionValidator.validateFileFormat('invalid-file-format'))
            .to.throw('Please provide valid input for option: file-format.'
                    + ' Value must be in (application/json, text/csv).');
        });

        validatorTests(optionValidator.validateFileFormat, 'file-format');
    });

    describe('# validate sortField', () => {
        it('| should allow valid format', () => {
            for (var i = 0; i < validSortFieldForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateSortField(validSortFieldForNLUEvaluation[i])).to.not.throw();
            }
        });

        it('| should expect strict casing', () => {
            for (var i = 0; i < validSortFieldForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateSortField(validSortFieldForNLUEvaluation[i].toLowerCase()))
                .to.throw('Please provide valid input for option: sort-field.'
                        + ' Value must be in (STATUS, ACTUAL_INTENT, EXPECTED_INTENT).');
            }
        });

        it('| should allow invalid format', () => {
            expect(() => optionValidator.validateSortField('invalid-sort-field'))
            .to.throw('Please provide valid input for option: sort-field.'
                    + ' Value must be in (STATUS, ACTUAL_INTENT, EXPECTED_INTENT).');
        });

        validatorTests(optionValidator.validateSortField, 'sort-field');
    });

    describe('# validate testCaseStatus', () => {
        it('| should allow valid format', () => {
            for (var i = 0; i < validTestCaseStatusForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateTestCaseStatus(validTestCaseStatusForNLUEvaluation[i])).to.not.throw();
            }
        });

        it('| should expect strict casing', () => {
            for (var i = 0; i < validTestCaseStatusForNLUEvaluation.length; i++) {
                expect(() => optionValidator.validateTestCaseStatus(validTestCaseStatusForNLUEvaluation[i].toLowerCase()))
                .to.throw('Please provide valid input for option: test-case-status.'
                        + ' Value must be in (PASSED, FAILED).');
            }
        });

        it('| should allow invalid format', () => {
            expect(() => optionValidator.validateTestCaseStatus('invalid-test-case-status'))
            .to.throw('Please provide valid input for option: test-case-status.'
                    + ' Value must be in (PASSED, FAILED).');
        });

        validatorTests(optionValidator.validateTestCaseStatus, 'test-case-status');
    });

    describe('# validate actualIntentName', () => {
        it('| should allow valid format', () => {
            expect(() => optionValidator.validateActualIntentName('actualIntentName')).to.not.throw();
        });

        validatorTests(optionValidator.validateActualIntentName, 'actual-intent-name');
    });

    describe('# validate expectedIntentName', () => {
        it('| should allow valid format', () => {
            expect(() => optionValidator.validateExpectedIntentName('expectedIntentName')).to.not.throw();
        });

        validatorTests(optionValidator.validateExpectedIntentName, 'expected-intent-name');
    });

    describe('# validate file exists', () => {
        it('| should allow valid file path', () => {
            fs.existsSync.returns(true);
            expect(() => optionValidator.validateFile('file-path')).to.not.throw();
        });

        it('| should not allow non existent file path', () => {
            fs.existsSync.returns(false);
            expect(() => optionValidator.validateFile('file-path'))
                .to.throw('File not exsist. Please verify file path is correct.');
        });

        validatorTests(optionValidator.validateFile, 'file');
    });

    function validatorTests(func, field) {
        it('| should not succeed on null or empty', () => {
            const valueEmptyErr = `Please provide valid input for option: ${field}. Value must not be empty.`;
            const requiredErr = `Please provide valid input for option: ${field}. Field is required and must be set.`;
            expect(() => func(null)).to.throw(requiredErr);
            expect(() => func()).to.throw(requiredErr);
            expect(() => func('')).to.throw(valueEmptyErr);
            expect(() => func('           ')).to.throw(valueEmptyErr);
        });

        it('| should override field name', () => {
            expect(() => func(null, { field: 'test' }))
                .to.throw('Please provide valid input for option: test. Field is required and must be set.');
        });

        it('| should succeed when not required', () => {
            expect(() => func(null, { required: false })).to.not.throw();
            expect(() => func(undefined, { required: false })).to.not.throw();
            expect(() => func('', { required: false })).to.throw();
            expect(() => func('   ', { required: false })).to.throw();
        });
    }
});
